import re
import time

from pylabs import q
from pylabs.baseclasses.CommandWrapper import CommandWrapper

from VirtualboxEnums import VirtualboxMachineStatusType, VirtualBoxDeviceType, VirtualboxHardDriveType,\
                            VirtualboxNetworkingMod, VirtualBoxSessionState, VirtualBoxAccessMode, VirtualBoxLockType,\
                            VirtualBoxCleanupModeType, VirtualBoxUSBDeviceState

from VBoxTools import getSessionObject, getVirtualBox,getMachineStateEnumFromStateNr, getSessionStateEnumFromStateNr, vbox as mgr
from VirtualBoxTools import VirtualBoxTools

import init

class VirtualBoxMachineConfiguration(CommandWrapper):
    """
    VirtualBox Machine configuration commands
    """
    ___ISCSI_HARD_DISK_FORMAT_                  = "iSCSI"
    ___HOST_IQN_SEPARATOR_                      = "|"
    ___SERVER_PORT_SEPARATOR_                   = ":"
    ___HARD_DISK_PROPERTY_TARGET_ADDRESS_       = "TargetAddress"
    ___HARD_DISK_PROPERTY_LUN_                  = "LUN"
    ___HARD_DISK_PROPERTY_USERNAME_             = "InitiatorUsername"
    ___HARD_DISK_PROPERTY_PASSWORD_             = "InitiatorSecret"
    ___HARD_DISK_PROPERTY_TARGET_NAME_          = "TargetName"
    ___HARD_DISK_PROPERTY_INITIATOR_NAME_       = "InitiatorName"
    ___HARD_DISK_PROPERTY_VALUE_INITIATOR_NAME_ = "iqn.2008-04.com.sun.virtualbox.initiator"
    ___SATA_STORAGE_CONTROLLER_                 = "SATA"
    ___ERROR_MESSAGE_INVALID_HARD_DISK_ORDER_   = "Invalid hard disk order: "


    def registerMachine(self, name, baseFolderPath = None, register = True, cpucount = 1, ioapic = False):
        """
        Create and register a new vbox machine, assigning it the given name and baseFolder

        @type name: string
        @param name: unique name for the new machine
        @type baseFolderPath: path
        @param baseFolderPath: name of the folder in which the XML definition file will be written, for example on linux platform this folder would be typically created under $HOME/.VirtualBox/
        @type register:boolean
        @param register: register the machine with the VBox server not only create the configuration file
        @type cpucount:integer
        @param cpucount: number of processors for this machine
        @type osbitversion: string
        @param osbitversion: BIT32 or BIT64
        """

        machine = None
        vbox = None

        if not name or not (name.isdigit() or q.system.fs.validateFilename(name) ):
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hyphens.')

        vbox = getVirtualBox()

        try:
            if not self.machineExists(name):
                machine          = vbox.createMachine(baseFolderPath, name, None, '00000000-0000-0000-0000-000000000000', False)
                machine.CPUCount = int(cpucount)
                if ioapic:
                    machine.BIOSSettings.IOAPICEnabled = ioapic
                controller = machine.addStorageController('SATA' , 2)
                controller.useHostIOCache            = False
                machine.audioAdapter.enabled         = True
                machine.audioAdapter.audioController = 0 # AC97
                machine.audioAdapter.audioDriver     = 3 # ALSA
                machine.saveSettings()
                # Now we should register the machine.
                if register:
                    vbox.registerMachine(machine)
                q.logger.log("Machine %s created successfully" %name, 3)
                return machine.id
        except Exception, e:
            if machine:
                # Machine created but failed to register
                machine.deleteSettings()
            q.logger.log("Failed to create Machine [%s].\n Exception: %s" % (name, e), 3)
            raise e

    def unregisterMachine(self, name, delete = True):
        """
        Unregister the machine specified by <name> from vbox

        @type name: string
        @param name: machine name
        @type delete: boolean
        @param delete: delete machine XML definition file will be deleted, default True
        """

        machine = None
        vbox = getVirtualBox()

        if not name or not (name.isdigit() or q.system.fs.validateFilename(name)) :
            raise ValueError('The name [%s] has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hyphens.'%name)

        q.logger.log("Unregistering machine %s" %name, 3)

        try:
            machine = self._getMachineObjByName(name)
        except:
            q.logger.log('Machine %s already unregisterd'%name, 3)
            return


        if delete:
            for order in xrange(vbox.systemProperties.networkAdapterCount):
                interface = machine.getNetworkAdapter(order)
                hostInterface = getattr(interface, 'hostInterface')
                if hostInterface and hostInterface in q.cmdtools.tunctl.listInterfaces():
                    q.cmdtools.tunctl.deleteTun(hostInterface)

            for disk in self.getHardDisksAttached(machine.name):
                if q.system.fs.exists(disk['location']):
                    q.system.fs.removeFile(disk['location'])
                if q.system.fs.isEmptyDir(q.system.fs.getDirName(disk['location'])):
                    q.system.fs.removeDir(q.system.fs.getDirName(disk['location']))

        mediums = machine.unregister(VirtualBoxCleanupModeType.FULL.value)

        if delete:
            machine.delete(mediums)

    def getMachineStatus(self, name):
        """
        Print the detailed status for the virtual machine specified by <name>

        @type name: string
        @param name: machine name

        @return: detailed status for the  virtual machine specified by <name>
        """
        if not name or not (name.isdigit() or q.system.fs.validateFilename(name)) :
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hypens.')

        try:
            machine = self._getMachineObjByName(name)
            return self._getMachineStatus(machine)
        except:
            return { 'State' : 'NotExisting'}

    def listMachines(self, byName=False):
        """
        Lists status of all vbox machines
        """

        machines = {}

        q.logger.log("Listing information of all registered machines", 3)

        vbox = getVirtualBox()
        for machine in vbox.getMachines():
            try:
                machines[machine.name if byName else str(machine.id)] = self._getMachineStatus(machine)
            except:
                pass

        return machines

    def getMachines(self):
        """
        @return: All available machine names.
        """
        vbox = getVirtualBox()
        return [machine.name for machine in vbox.getMachines()]

    def machineExists(self, name):
        """
        @param name: Machine name
        @return: True, if machine exists, False otherwise
        """
        try:
            self._getMachineObjByName(name)
            return True
        except Exception, e:
            return False

    def addDisk(self, machineName, port, path, asyncIO = False):
        """
        Add disk to virtual box machine

        @param machineName: unique name for the machine
        @param controllerPort: Sequence of the disk, number ranging from 0 to 29.
        @param path: Path to the disk image on the host
        @param asyncIO: enable async IO on disk
        """

        machine = None
        session = None
        device = 0 # For SATA and SCSI device always needs to be 0
        vbox = None

        if not machineName or not (machineName.isdigit() or q.system.fs.validateFilename(machineName)) :
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hypens.')

        if not q.system.fs.isFile(path):
            raise RuntimeError("File '%s' not found" % path)

        vbox = getVirtualBox()

        machine = self._getMachineObjByName(machineName)
        if getMachineStateEnumFromStateNr(machine.state) == VirtualboxMachineStatusType.RUNNING:
            raise RuntimeError("Machine '%s' is running" %machineName)

        try:
            disk = vbox.findMedium(path)
        except Exception, e:
            # Register hard disk
            disk = vbox.openMedium(path, VirtualBoxDeviceType.HARDDISK, VirtualBoxAccessMode.READONLY.value)

        controller = self._getStorageController(machine)
        attachedDisks = self._getAttachedDisks(machine, controller)

        for attachedDisk in attachedDisks:
            if attachedDisk.id == disk.id:
                raise RuntimeError("Disk '%s' already attached to machine '%s'" % (path, machineName))

        session,machine = self._openSession(machine.name)

        q.logger.log("Adding a disk with order %(order)s for machine %(machine)s" %{'order': port, 'machine': machineName}, 3)

        # When the session is opened, use the session.machine object instead of machine
        # since session.machine object is mutable
        try:
            """
            For an SATA controller, controllerPort must be a number ranging from 0 to 29.
            For a SCSI controller, controllerPort must be a number ranging from 0 to 15.
            For both SCSI and SATA, the device parameter is unused and must be 0.
            """
            if asyncIO:
                #Enabled async I/O [VBox 4.0.10r72479]
                session.machine.setExtraData('VBoxInternal/Devices/ahci/0/LUN#%d/AttachedDriver/Config/UseNewIo'%port, '1')

            session.machine.attachDevice(controller.name, port, device, VirtualBoxDeviceType.HARDDISK.value, disk)
            session.machine.saveSettings()
            q.logger.log("Disk hd%(order)s added to machine[%(machine)s]successfully" %{'order': port, 'machine': machineName}, 3)
        finally:
                self._closeSession(session)

    def getHardDisksAttached(self, machineName, busType='SATA'):
        """
        Retrieve a detailed list of hard disks attached to given machine
        @param machineName: name of the machine to retrieve disks for
        """
        machine = self._getMachineObjByName(machineName)
        return [medium for medium in self.__getMediumAttachments(machine) if medium['type'] == str(VirtualBoxDeviceType.HARDDISK)]

    def showvminfo(self, machineName):
        """
        Show detailed info of given machine
        @param machineName: name of the machine to show details for
        """
        vbox = getVirtualBox()
        mach = self._getMachineObjByName(machineName)
        info = dict()
        os   = vbox.getGuestOSType(mach.OSTypeId)
        gps  = VirtualBoxTools().listGuestProperties(machineName)
        info['ostype']  = os.description
        properties      = ['name', 'id', 'CPUCount', 'memorySize', 'VRAMSize', 'monitorCount', 'clipboardMode', 'accelerate3DEnabled', 'accelerate2DVideoEnabled']
        for prop in properties:
            info[prop]  = getattr(mach, prop)

        if mach.sessionPid == 4294967295:
            info['sessionPid'] = 0
        else:
            info['sessionPid'] = mach.sessionPid
        bios = mach.BIOSSettings
        import time
        ports                       = mach.VRDEServer.getVRDEProperty('TCP/Ports')
        info['lastChanged']         = time.asctime(time.localtime(long(mach.lastStateChange)/1000))
        info['VRDP']                = {'enabled': mach.VRDEServer.enabled,
                                       'ports'  : [int(port) for port in ports.split(',') if port not in ('', None)] if ports else None}
        info['bios']                = {'ACPIEnabled': bios.ACPIEnabled, 'IOAPICEnabled': bios.IOAPICEnabled}
        info['firmwareType']        = self.__getFirmwareType(mach.firmwareType)
        info['hwVirtEnabled']       = mach.getHWVirtExProperty(mgr.constants.HWVirtExPropertyType_Enabled)
        info['hwVirtVPID']          =  mach.getHWVirtExProperty(mgr.constants.HWVirtExPropertyType_VPID)
        info['hwVirtNestedPaging']  = mach.getHWVirtExProperty(mgr.constants.HWVirtExPropertyType_NestedPaging)
        info['state']               = self.__getMachineState(mach.state)
        info['storagecontrollers']  = self.__getStorageControllers(mach)
        info['mediums']             = self.__getMediumAttachments(mach)
        info['networkAdapters']     = list()
        for order in xrange(vbox.systemProperties.networkAdapterCount):
            adapter     = mach.getNetworkAdapter(order)
            properties  = ['slot', 'enabled', 'MACAddress', 'hostInterface', 'internalNetwork', 'NATNetwork', 'cableConnected', 'lineSpeed', 'traceEnabled']
            details     = dict()
            for prop in properties:
                details[prop] = getattr(adapter, prop)

            gip = [gp['value'] for gp in gps if re.match('^\/VirtualBox\/GuestInfo\/Net\/%s\/V4\/IP$'%order, gp['name'])]
            details['ipaddress'] = gip[0] if gip else ''

            details['adapterType']      = self.__getNetworkAdapterType(adapter.adapterType)
            details['attachmentType']   = self.__getNetworkAttachmentType(adapter.attachmentType)
            info['networkAdapters'].append(details)

        #Get Remotedisplayinfo
        if getMachineStateEnumFromStateNr(mach.state) == VirtualboxMachineStatusType.RUNNING:
            session,mach = self._openSession(mach.name)
            try:
                info['remoteDisplayInfo'] = {'clientName'   : session.console.VRDEServerInfo.clientName,
                                             'clientIP'     : session.console.VRDEServerInfo.clientIP}
            finally:
                self._closeSession(session)
        else:
            info['remoteDisplayInfo'] = {'clientName':'', 'clientIP':''}

        return info


    def queryMetrics(self,  machineNames, metrics=['*']):
        """
        query metrics of machine
        @param machineNames: list of machine names
        @param metrics: list of metrics to query e.g. [RAM/Usage/Used]
        """
        vbox        = getVirtualBox()
        perf        = mgr.getPerfCollector(vbox)
        machines    = list()
        if not isinstance(machineNames, (list,tuple)):
            machineNames=[machineNames]

        for machineName in machineNames:
            machines.append(self._getMachineObjByName(machineName))
        return perf.query(metrics, machines)

    def setupMetrics(self, period, count, machineNames=[], metrics=['*']):
        """
        Sets parameters of specified base metrics for a set of machines
        @param machineNames: list of machine names
        @param metrics: list of metrics to set e.g. [RAM/Usage/Used]
        @param period: time interval between samples
        @param count: number of samples to keep
        """
        vbox        = getVirtualBox()
        perf        =      mgr.getPerfCollector(vbox)
        machines    = list()
        if not isinstance(machineNames, (list,tuple)) and machineNames:
            machineNames=[machineNames]

        for machineName in machineNames:
            machines.append(self._getMachineObjByName(machineName))
        return perf.setup(metrics, machines, period, count)

    def removeDisk(self, machineName, controllerPort, delete = False, asyncIO=True):
        """
        Remove disk from virtual box machine

        @param machineName: unique name for the machine
        @param controllerPort: sequence of the disk, for SATA: a number ranging from 0 to 29.
        @param asyncIO: async IO enabled on disk level
        """

        deviceToDetach  = None
        machine         = None
        session         = None
        vbox            = None
        port            = controllerPort
        device          = 0 # for SATA and SCSI the device number is unused and should always be 0

        if not machineName or not (machineName.isdigit() or q.system.fs.validateFilename(machineName)) :
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hyphens.')

        q.logger.log("Unregistering disk hd%(order)s images from machine[%(machine)s] " %{'order': controllerPort, 'machine': machineName}, 3)

        vbox = getVirtualBox()

        try:
            machine = self._getMachineObjByName(machineName)
        except:
            q.logger.log('Machine %s already removed, skipping remove disk on controllerport %s'%(machineName,controllerPort), 3)
            return

        if getMachineStateEnumFromStateNr(machine.state) == VirtualboxMachineStatusType.RUNNING:
            raise RuntimeError("Machine '%s' is running" %machineName)

        controller      = self._getStorageController(machine)
        attachedDevices = self._getAttachedDevices(machine, controller)

        for attachedDevice in attachedDevices:
            if attachedDevice.port == port and attachedDevice.device == device:
                deviceToDetach = attachedDevice.medium

        if not deviceToDetach:
            q.logger.log("No disk attached on machine '%s' with order '%s'" %(machineName, controllerPort),3)
            return

        session,machine = self._openSession(machine.name)

        try:
            session.machine.detachDevice(controller.name, port, device)
            session.machine.saveSettings()
            q.logger.log("Disk hd%(order)s removed successfully from machine[%(machine)s]successfully"%{'order':controllerPort, 'machine':machineName}, 3)
        finally:
            self._closeSession(session)

        # Unregister the disk from the virtual box system

        deviceToDetach.close()

        # If delete flag is on, delete the HDD physically
        if delete and hasattr(deviceToDetach,'location'):
            if q.system.fs.exists(deviceToDetach.location):
                q.logger.log("Removing HDD with location: %s" %deviceToDetach.location, 3)
                q.system.fs.removeFile(deviceToDetach.location)
                q.logger.log("Successfully removed HDD in location: %s" %deviceToDetach.location, 3)

    def addNic(self, machineName, order, networkingMod = q.enumerators.VirtualboxNetworkingMod.NAT,
               nicType = q.enumerators.VirtualboxNicType.AM79C973, hostInterfaceDevice = None, macAddress = None):
        """
        Add NIC to virtual box machine

        @param machineName: unique name for the machine
        @param order: sequence of the NIC on the machine
        @type networkingMod: VirtualboxNetworkingMod
        @param networkingMod: the type of the networking architecture that will be virtualized, on of the value under q.enumerators.VirtualboxNetworkingMod.[TAB], by default it would be set to q.enumerators.VirtualboxNetworkingMod.NAT
        @type nicType: VirtualboxNicType
        @param nicType: type of the network interface card, one of the values of q.enumerators.VirtualboxNicType
        @param hostInterfaceDevice: name of the virtual network interface on the host system, provided when the networking mod is host interface networking
        @param macAddress: the network device's MAC address
        """

        nic     = None
        vbox    = None
        session = None

        if not machineName or not (machineName.isdigit() or q.system.fs.validateFilename(machineName)) :
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hyphens.')

        if not q.platform.isSolaris() and networkingMod == q.enumerators.VirtualboxNetworkingMod.HOSTIF and not hostInterfaceDevice:
            raise ValueError('Host Interface Networking Mode requires host interface device to be supplied')

        vbox    = getVirtualBox()
        order   = int(order)

        if order >= int(vbox.systemProperties.networkAdapterCount):
            raise ValueError("Order must be between '0' and '%s'" % vbox.systemProperties.networkAdapterCount)

        machine = self._getMachineObjByName(machineName)

        if getMachineStateEnumFromStateNr(machine.state) == VirtualboxMachineStatusType.RUNNING:
            raise RuntimeError("Machine '%s' is running" %machineName)

        session,machine = self._openSession(machine.name)

        nic = session.machine.getNetworkAdapter(order)
        if macAddress == 'auto':
            macAddress = None

        try:
            if networkingMod == q.enumerators.VirtualboxNetworkingMod.NOTATTACHED:
                nic.enabled     = False
                nic.detach()
            else:
                nic.adapterType = nicType.value
                nic.MACAddress  = macAddress
                nic.enabled     = True
                if networkingMod == q.enumerators.VirtualboxNetworkingMod.NAT:
                    nic.attachToNAT()
                elif networkingMod == q.enumerators.VirtualboxNetworkingMod.INTNET:
                    nic.attachToInternalNetwork()
                elif networkingMod == q.enumerators.VirtualboxNetworkingMod.BRIDGED:
                    if hostInterfaceDevice:
                        nic.hostInterface = hostInterfaceDevice
                    nic.attachToBridgedInterface()
                elif networkingMod == q.enumerators.VirtualboxNetworkingMod.HOSTIF:
                    if hostInterfaceDevice:
                        nic.hostInterface = hostInterfaceDevice
                    nic.attachToHostOnlyInterface()
                else:
                    raise RuntimeError("Unknown networking mode supplied")

            session.machine.saveSettings()
            q.logger.log("NIC nic%(order)s modified successfully to machine[%(machine)s]successfully"%{'order'  : order,
                                                                                                       'machine': machineName}, 3)
        finally:
            self._closeSession(session)


    def removeNic(self, machineName, order):
        """
        Remove/Detach the NIC from a VMachine

        @param machineName: unique name for the machine
        @param order: sequence of the nic on the machine
        """

        self.addNic(machineName, order, q.enumerators.VirtualboxNetworkingMod.NOTATTACHED)


    def removeISCSITarget(self, uuid):
        """
        Unregisters an iscsi target from the VBox instance

        @param uuid: uuid representing an identifier to an iscsi target attached to VBox instance
        """
        # IHardDisk instance by which we are able to control the registered iSCSI targetS
        hardDisk = None
        vbox = getVirtualBox()

        # Get the IHardDisk instance using hard disk uuid
        hardDisk = vbox.getHardDisk(uuid)
        hardDisk.close()

    def addISCSIDisk(self,machineName, server, targetName, port = None, lun = None, userName = None, password = None):
        """
        Registers an iscsi target as a disk in VBox

        @type server: string or q.enumerators.ipaddress
        @param server: domain name or ip address of the server where the iscsi target located
        @type targetName: string
        @param targetName: unique name of the target in the server
        @type port:integer
        @param port: port number on the server machine that will be connected to for the iscsi target
        @type lun: integer
        @param lun: Logical Unit Number of the target resource
        @param userName: login of the iscsi target server
        @param password: password of the iscsi target server
        @rtype: string
        @return: the uuid of the created iscsi disk
        """
        vbox        = getVirtualBox()
        name        = server + self.___HOST_IQN_SEPARATOR_ + targetName
        hdds        = vbox.getHardDisks()
        hdds        = [hdd for hdd in hdds if hdd.name == name]
        machine     = None
        session     = None
        device      = 0 # For SATA and SCSI device always needs to be 0

        if not machineName or not (machineName.isdigit() or q.system.fs.validateFilename(machineName)) :
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hypens.')

        machine     = self._getMachineObjByName(machineName)
        if getMachineStateEnumFromStateNr(machine.state) == VirtualboxMachineStatusType.RUNNING:
            raise RuntimeError("Machine '%s' is running" %machineName)

        try:

            controller      = self._getStorageController(machine = machine,
                                                         busType = 'SCSI')
            if not hdds:
                q.hypervisors.cmdtools.virtualbox.tools.addISCSIDisk(machineName    = machineName,
                                                                     controllername = controller.name,
                                                                     port           = port,
                                                                     server         = server,
                                                                     target         = targetName)
                hdds        = vbox.getHardDisks()
                hdds        = [hdd for hdd in hdds if hdd.name == name]

            disk = hdds[0]

            q.logger.log("Adding a disk with order %(order)s for machine %(machine)s" %{'order': port, 'machine': machineName}, 3)

            return disk
        finally:
            if session:
                self._closeSession(session)

    def updateMachine(self, machineName, os = None, memory = None, dvd = None, kvmport = None,
                      hwvirtualization = None, pae = None, cpucount = None,\
                      vram=None, usb = None, monitors = None, cpuexecutioncap = None):
        """
        Update machine parameters, os, memory, dvd, ... etc

        @type machineName:strin
        @param machineName: updated machine name for the machine that would be update
        @type os:VirtualboxOsType
        @param os:Type of operating system running on the machine, one of the enumerated values of q.enumerators.VirtualboxOsType.
        @type memory:integer
        @param memory: RAM in MB configured on the machine
        @type dvd:path
        @param dvd: dvd/image attached to the machine (image uuid)
        @type kvmport:integer
        @param kvmport: Port number assigned for kvm, used for remote access of the vbox process
        @type hwvirtualization:boolean
        @param hwvirtualization: Indicates if the hardware virtualization extensions are enabled
        @type pae: boolean
        @param pae: Indicates if the physical address extension are enabled
        @param usb: Indicates if the usb support is enabled
        @param monitors: dict of monitors configuration
        @param cpuexecutioncap: percentage to limit the number of CPU cycles (CPU cycles per second) on the host
        """

        if not machineName or not (machineName.isdigit() or q.system.fs.validateFilename(machineName) ):
            raise ValueError('The name has an invalid format. It must start with a character followed by alphanumeric characters, underscores or hypens.')

        vbox = getVirtualBox()

        machine = self._getMachineObjByName(machineName)

        if machine.state == int(q.enumerators.VirtualboxMachineStatusType.RUNNING):
            raise RuntimeError("Machine '%s' is running" % machineName)

        session,machine = self._openSession(machine.name)

        try:
            if os:
                q.logger.log("Processing updated operating system configuration", 3)
                session.machine.OSTypeId = os.value
            if memory:
                q.logger.log("Processing updated memory configuration", 3)
                session.machine.memorySize = memory

            if dvd:
                imgUuid = None

                if q.system.fs.isFile(dvd):
                    # DVD is an image file
                    dvdimage = None
                    try:
                        dvdimage = vbox.findDVDImage(dvd)
                    except Exception, e:
                        # Can't find image, instead open it.
                        dvdimage = vbox.openDVDImage(dvd, '00000000-0000-0000-0000-000000000000')

                    imgUuid = dvdimage.id
                else:
                    imgUuid = dvd


                #q.logger.log("Processing updated dvd configuration", 3)
                #session.machine.attachDevice('DVD', 0, 0, VirtualBoxDeviceType.DVD, imgUuid)

                #session.machine.mountMedium("DVD", '', '', imgUuid, False)
                #session.machine.DVDDrive.mountImage(imgUuid)

            if kvmport:
                q.logger.log("Processing updated kvmport configuration", 3)
                session.machine.VRDEServer.setVRDEProperty("TCP/Ports", str(kvmport))
                session.machine.VRDEServer.enabled = True
                session.machine.VRDEServer.setVRDEProperty("VideoChannel/Enabled", "true")
            if monitors:
                monitorCount = len(monitors)
                session.machine.VRDEServer.allowMultiConnection = monitorCount > 1
                session.machine.monitorCount = monitorCount

            if not hwvirtualization is None:
                q.logger.log("Processing updated hardware virtualization configuration", 3)
                session.machine.HWVirtExEnabled = hwvirtualization

            if not pae is None:
                q.logger.log("Processing updated PAE configuration", 3)
                session.machine.PAEEnabled = pae

            if cpucount is not None:
                if cpucount > 1:
                    session.machine.BIOSSettings.IOAPICEnabled = True
                session.machine.CPUCount = int(cpucount)

            if usb is not None:
                session.machine.USBController.enabled = usb
                session.machine.USBController.enabledEhci = True

            if cpuexecutioncap is not None:
                session.machine.CPUExecutionCap = cpuexecutioncap

            #set vram to recommended value
            #@TODO: vram should be configurable
            self._setVRAM(vbox, session.machine, vram)

            session.machine.saveSettings()
        finally:
            try:
                self._closeSession(session)
            except Exception, e:
                q.logger.log("Error while closing VBox session. Message: %(msg)s" %{'msg':str(e)}, 3)


    def _setVRAM(self, vbox, machine, vram=None):
        recommendedVRAM = vbox.getGuestOSType(machine.OSTypeId).recommendedVRAM
        if not vram:
            if machine.VRAMSize < recommendedVRAM:
                machine.VRAMSize = recommendedVRAM

        elif vram < recommendedVRAM:
            raise ValueError('VRAM %s is less than recommended value for GuestOS %s'%(vram,machine.OSTypeId))
        else:
            machine.VRAMSize = vram

    def vmStatistics(self, machineName, pattern='*', showDescriptions=True):
        """
        Get the statistics of given machine in a XML format.
        @param machineName: name of the machine
        @param pattern: pattern matching the name e.g. /Device/*, /PGM/Page/cWrittenToPages, ..
        """
        machine = self._getMachineObjByName(machineName)
        session,machine = self._openSession(machineName)
        try:
            stats = session.console.debugger.getStats(pattern, showDescriptions)
        finally:
            self._closeSession(session)
        return stats

    def getVideoMode(self, machineName, monitorOrder):
        """
        Get the current video mode configuration of given machine

        @param machineName: name of the machine
        @param monitorOrder: order of the monitor
        @return width,height,bpp
        """
        machine = self._getMachineObjByName(machineName)
        session,machine = self._openSession(machineName)
        try:
            videoMode = session.console.display.getScreenResolution(monitorOrder)
        finally:
            self._closeSession(session)
        return videoMode

    def setVideoMode(self, machinename, monitorOrder, width, height, bitsPerPixel):
        """
        Set new video mode configuration for the given machine

        @param machinename: name of the vmachine
        @param monitorOrder: order of the monitor
        @param width: width of the screen
        @param height: height of the screen
        @param bitsPerPixel: BPP that needs to be used [16/24/32]
        """
        machine = self._getMachineObjByName(machinename)
        session,machine = self._openSession(machinename)
        try:
            session.console.display.setVideoModeHint(width if width else 1024,
                                                     height if height else 768,
                                                     bitsPerPixel if bitsPerPixel else 24,
                                                     monitorOrder)
        except: pass
        finally:
            self._closeSession(session)

    def setMonitorCount(self, machineName, monitorCount):
        """
        Set amount of monitors for the specified machine

        @param machineName: name of the vmachine
        @param monitorCount: total numbers of monitors
        """

        machine = self._getMachineObjByName(machineName)

        if machine.state == int(q.enumerators.VirtualboxMachineStatusType.RUNNING):
            raise RuntimeError("Machine '%s' is running" % machineName)

        session,machine = self._openSession(machineName)

        try:
            session.machine.monitorCount                    = monitorCount
            session.machine.VRDEServer.allowMultiConnection = monitorCount > 1
            session.machine.saveSettings()
        finally:
            try:
                self._closeSession(session)
            except Exception, e:
                q.logger.log("Error while closing VBox session. Message: %(msg)s" %{"msg": str(e)}, 3)

    def addUSBDevice(self, machineName, usbdevice):
        q.logger.log("Adding USB device(s)", 3)
        machine         = self._getMachineObjByName(machineName)
        session,machine = self._openSession(machineName)

        try:
            USBDevices = VirtualBoxTools.listHostUsbDevices()
            availableUSBDevices = [usbdev.id for usbdev in USBDevices if usbdev.state in (VirtualBoxUSBDeviceState.AVAILABLE.value,
                                                                                          VirtualBoxUSBDeviceState.BUSY.value)]
            if usbdevice.id in availableUSBDevices:
                session.console.attachUSBDevice(usbdevice.id)
        finally:
            self._closeSession(session)

    def removeUSBDevice(self, machineName, usbdevice):
        q.logger.log("Removing USB device(s)", 3)
        machine         = self._getMachineObjByName(machineName)
        session,machine = self._openSession(machineName)

        try:
            session.console.detachUSBDevice(usbdevice.id)
        finally:
            self._closeSession(session)

    def _openSession(self, machineName):
        vbox    = getVirtualBox()
        machine = self._getMachineObjByName(machineName)
        session = getSessionObject()
        machine.lockMachine(session, VirtualBoxLockType.SHARED.value)
        return session, machine

    def _closeSession(self, session):
        try:
            session.unlockMachine()
        except Exception, e:
            q.logger.log("Error while closing VBox session. Error: %s" %e, 3)

    def _getMachineObjByName(self, machineName):
        vbox = getVirtualBox()
        try:
            machine = vbox.findMachine(machineName)
            return machine
        except Exception, e:
            #Find by UUID
            machines = self.listMachines(byName=True)
            if machineName in machines:
                return vbox.findMachine(machines[machineName]['UUID'])
            raise ValueError("Can't find machine %s" % machineName)

    def _getStorageController(self, machine, busType = 'SATA'):
        controller = [st for st in machine.getStorageControllers() if st.bus == mgr.constants._Values['StorageBus'][busType]]
        if not controller: raise ValueError('No StorageController of bus type %s attached to machine %s'%(busType, machine.name))
        return controller[0]

    def _getAttachedDevices(self, machine, controller):
        attachedMediums = machine.getMediumAttachmentsOfController(controller.name)
        return attachedMediums

    def _getAttachedDisks(self, machine, controller):
        return [attached.medium for attached in self._getAttachedDevices(machine, controller) if int(attached.type) == int(VirtualBoxDeviceType.HARDDISK)]

    def __getMachineState(self, machineState):
        for key, value in mgr.constants._Values["MachineState"].iteritems():
            if int(value) == int(machineState):return key

    def _getMachineStatus(self, machine):
        data                = dict()
        ports               = machine.VRDEServer.getVRDEProperty('TCP/Ports')
        data['Name']        = machine.name
        data['Guest OS']    = machine.OSTypeId
        data['UUID']        = str(machine.id)
        data['Config File'] = machine.settingsFilePath
        data['Memory Size'] = machine.memorySize
        data['VRAM Size']   = machine.VRAMSize
        data['State']       = self.__getMachineState(machine.state)
        data['VRDP Ports']  = [int(port) for port in ports.split(',') if port not in ('', None)] if ports else None
        return data

    def __getMediumState(self, mediumState):
        for key, value in mgr.constants._Values["MediumState"].iteritems():
            if int(value) == int(mediumState):return key

    def __getMediumAttachments(self, machine):
        attaches = mgr.getArray(machine, 'mediumAttachments')
        mediums  = list()
        for attach in attaches:
            medium = attach.medium
            details = dict()
            for prop in ['id', 'logicalSize', 'location', 'name', 'format', 'size', 'autoReset', 'lastAccessError', 'readOnly', 'hostDrive', 'description']:
                details[prop] = getattr(medium, prop, '')
            details['state'] = self.__getMediumState(medium.state)
            details['type'] = self.__getMediumType(medium.deviceType)
            details['size'] = medium.size
            details['controller'] = attach.controller
            mediums.append(details)
        return mediums

    def __getStorageControllers(self, machine):
        controllers = list()
        for controller in machine.getStorageControllers():
            details = dict()
            details['name'] = controller.name
            details['type'] = self.__getControllerType(controller.controllerType)
            details['bus'] = self.__getStorageBus(controller.bus)
            controllers.append(details)
        return controllers

    def __getStorageBus(self, storageBus):
        for key, value in mgr.constants._Values['StorageBus'].iteritems():
            if int(value) == int(storageBus):return key

    def __getControllerType(self, controllerType):
        for key, value in mgr.constants._Values['StorageControllerType'].iteritems():
            if int(value) == int(controllerType):return key

    def __getMediumType(self, mediumType):
        for key, value in mgr.constants._Values['DeviceType'].iteritems():
            if int(value) == int(mediumType):return key

    def __getFirmwareType(self, firmwaretype):
        for key, value in mgr.constants._Values['FirmwareType'].iteritems():
            if int(value) == int(firmwaretype):return key

    def __getMachineState(self, machinestate):
        for key, value in mgr.constants._Values['MachineState'].iteritems():
            if int(value) == int(machinestate):return key

    def __getNetworkAdapterType(self, networkAdapterType):
        for key, value in mgr.constants._Values['NetworkAdapterType'].iteritems():
            if int(value) == int(networkAdapterType):return key

    def __getNetworkAttachmentType(self, networkAttachmentType):
        for key, value in mgr.constants._Values['NetworkAttachmentType'].iteritems():
            if int(value) == int(networkAttachmentType):return key


