from pylabs import q
from pylabs.baseclasses.CommandWrapper import CommandWrapper

from VirtualboxEnums import VirtualboxMachineStatusType, VirtualBoxLockType,\
                            VirtualBoxFrontEndType, VirtualBoxSessionStateType

import init

import time

from VBoxTools import getSessionObject
from VBoxTools import getVirtualBox
from VBoxTools import getMachineStateEnumFromStateNr

class VirtualBoxMachineControl(CommandWrapper):
    """
    VirtualBox Machine management commands, start, stop, reboot, pause, resume ... etc
    """

    # The errno of the exception thrown when trying to open VBox session while there is an open one already.
    # See the VBox SDK documentation.
    ___SESSION_ALREADY_OPENED_ERR_NO_ = -2135228409
    
    # Using this value with IProgress::waitForCompletion method makes the IProgress instance wait indefinitely till the whole process.
    # completes. 
    ___WAIT_FOR_INDEFINITE_TIME_      = -1
    
    # Max. wait time in milliseconds.
    ___MAX_WAIT_TIME_                  = 300000
    
    # IProgress::percent value which represents the total completion of the current process.
    ___100_PERCENT_PROGRESS_          = 100
    
    # Max. timeout time in seconds
    ___MAX_TIMEOUT_                   = 3600
    
    def start(self, name, kvm=None, monitors= dict(), compressionrate=0.75):
        """ 
        Start the machine.

        @param name: unique name for the machine.
        @param kvm: if kvm is set start machine in vdrp session.
        @param monitors: monitors of the machine
        @param compressionrate: compression rate [0.0->1.0]
        """

        # IMachine represents the VMachine we are trying to start.
        machine  = None
        
        # IProgress used with long running processes.
        progress = None
        
        # Mode into which we want to start the VMachine.
        mode     = None
        
        # IVirtualBox represents the VBox service and with which we open new sessions to start controlling VMachines.
        vbox     = getVirtualBox()
        
        #make certain that monitors is a dict
        monitors = dict() if monitors == None else monitors
        
        # Configure the VMachine execution mode.
        #
        if not kvm:
            mode = VirtualBoxFrontEndType.GUI.value
        else:
            mode = VirtualBoxFrontEndType.HEADLESS.value
        
        # Find the machine instance using its name.
        machine = vbox.findMachine(name)
        
        if self.__getMachineState(machine) == VirtualboxMachineStatusType.RUNNING:
            # Machine is running ?
            q.console.echo("Machine is running")
            return
        elif self.__getMachineState(machine) == VirtualboxMachineStatusType.PAUSED:
            q.console.echo("Machine is already paused. Please use 'resume' instead.")
            return

        # Open a remote session with the VMachine to be able to control the VMachine starting it and start it.
        session = None
        try:
            # Opening session and calling the IProgress::waitForCompletion causes the VMachine to start automatically
            # so we don't need to call IConsole::powerUp().
            session,machine = self._openSession(machineName = name,
                                                requireLock = False)
            progress = machine.launchVMProcess(session, mode, None)
            self.__watchProgressStatus(progress, self.___WAIT_FOR_INDEFINITE_TIME_)
            # Give VBOX some breathing space, to build the console
            time.sleep(1)
            for monitor in monitors:
                session.console.display.setVideoModeHint(monitor.width if monitor.width else 1024,
                                                         monitor.height if monitor.height else 768,
                                                         monitor.bpp if monitor.bpp else 24,
                                                         monitor.order)
            
            session.machine.VRDEServer.allowMultiConnection = len(monitors) > 1
            session.machine.VRDEServer.enabled = True
            session.machine.VRDEServer.setVRDEProperty("VideoChannel/Enabled", "true")
            
            if compressionrate and compressionrate > 0.0 and compressionrate <= 1.0:
                session.machine.VRDEServer.setVRDEProperty("VideoChannel/Quality", str(compressionrate*100))
            else:
                session.machine.VRDEServer.enabled = False
                session.machine.VRDEServer.setVRDEProperty("VideoChannel/Enabled", "false")
                session.machine.VRDEServer.setVRDEProperty("VideoChannel/Quality", "100")
        finally:
            # Close the session as we don't need it anymore.
            # Note: The locally opened session which was created by the VBox service while calling the IVirtualBox::openRemoteSession
            #       will still be opened.
            self._closeSession(session)
            
            
    def stop(self, name, forcestop=False, waitforstop=False):
        """
        Stop the machine.

        @param name: unique name for the machine.
        @type name: String.
        @param forcestop: Stop machine hard.
        @type forcestop: Bool.
        @param waitforstop: wait for machine to stop before returning max wait is 300second after this force stop is used.
        @type waitforstop: Bool.
        """

        # IConsole instance using which we can control the VMachine to stop.
        console = None        
        session = None
        
        # IVirtualBox represents the VBox service and with which we open new sessions to start controlling VMachines.
        vbox = getVirtualBox()
        
        # IMachine represents the VMachine we are trying to start.
        # Find the machine instance using its name.
        machine = vbox.findMachine(name)
        
        if self.__getMachineState(machine) == VirtualboxMachineStatusType.STOPPED:
            q.logger.log("Machine is already powered-off",1)
            return
        elif self.__getMachineState(machine) == VirtualboxMachineStatusType.PAUSED:
            forcestop = True
            waitforstop = False
            q.logger.log("Machine is paused, forcing stop",1)

        try:
            # Open the existing locally opened session by the VBox service.
            session,machine = self._openSession(machineName = name,
                                                requireLock = True)
            # Get the IConsole instance attached to this session so you can control the VMachine.
            console = session.console
            progress = None
            if not forcestop:
                q.logger.log('Calling powerbutton',1)
                # If not to force powering off the VMachine.
                # Send the ACPI PowerButton event.
                console.powerButton()
                # If the event was not handled by the VMachine's guest operating system.
                if not console.getPowerButtonHandled():
                    # @deprecated: IConsole::powerDownAsync should be used instead in later versions of VBox.
                    # @see: IConsole::powerDown and IConsole::powerDownAsync.
                    #
                    # For the VMachine to power off.
                    q.logger.log('Calling powerDown',1)
                    progress = console.powerDown()
            else:
                # Otherwise simply power off the VMachine.
                q.logger.log('Calling powerDown',1)
                progress = console.powerDown()
                
            if progress and waitforstop:
                try:
                    self.__watchProgressStatus(progress, self.___MAX_WAIT_TIME_)
                except:
                    progress = console.powerDown()
                    self.__watchProgressStatus(progress, self.___MAX_WAIT_TIME_)
            elif waitforstop:
                shutdownStartTime = time.time()
                while self.__getMachineState(machine) <> VirtualboxMachineStatusType.STOPPED:
                    if (time.time() - shutdownStartTime) >= self.___MAX_TIMEOUT_:
                        q.logger.log("Unable to poweroff machine after %s seconds."%self.___MAX_TIMEOUT_)
                        break
                    q.logger.log("Machine is still running, waiting for 5 seconds...")
                    time.sleep(5)
        finally:
            # Release the session
            self._closeSession(session)

    def reboot(self, name):
        """
        Reboot the machine.

        @param name: unique name for the machine.
        """
        
        # IVirtualBox represents the VBox service and with which we open new sessions to start controlling VMachines.
        vbox = getVirtualBox()
        
        # ISession instance to start controlling VMachines.
        
        # IMachine represents the VMachine we are trying to start.
        # Find the machine instance using its name.
        machine = vbox.findMachine(name)
        
        try:
            # Open the already existing session and reboot the VMachine
            session,machine = self._openSession(name)
            session.console.reset()
        finally:
            # Release the session
            self._closeSession(session)


    def pause(self, name):
        """
        Pause the machine.

        @param name: unique name for the machine.
        """

        # IVirtualBox represents the VBox service and with which we open new sessions to start controlling VMachines.
        vbox = getVirtualBox()
        
        # Find the machine instance using its name.
        # IMachine represents the VMachine we are trying to start.
        machine = vbox.findMachine(name)
        
        # Machine is already paused ?
        if self.__getMachineState(machine) == VirtualboxMachineStatusType.PAUSED:
            q.console.echo("Machine is already paused")
            return
        elif self.__getMachineState(machine) == VirtualboxMachineStatusType.STOPPED:
            q.console.echo("Can not pause an already powered-off machine")
            return
        
        try:
            # Open the already existing session and pause the VMachine
            session,machine = self._openSession(name)
            session.console.pause()
        finally:
            # Release the session
            self._closeSession(session)


    def resume(self, name):
        """
        Resume the machine if it's paused.

        @param name: unique name for the machine.
        """
        
        # IVirtualBox represents the VBox service and with which we open new sessions to start controlling VMachines.
        vbox = getVirtualBox()
        
        # IMachine represents the VMachine we are trying to start.
        # Find the machine instance using its name.
        machine = vbox.findMachine(name)
        
        if self.__getMachineState(machine) == VirtualboxMachineStatusType.RUNNING:
            # Machine is already running ?
            q.console.echo("Machine is already running")
            return
        elif self.__getMachineState(machine) == VirtualboxMachineStatusType.STOPPED:
            q.console.echo("Can not resume an already powered-off machine")
            return
        
        try:
            # Open the already existing session and resume the VMachine
            session,machine = self._openSession(name)
            session.console.resume()
        finally:
            self._closeSession(session)

    def state(self, name, unknownOnNotfound = True):
        """
        Get the real machine status from vbox.
        @param name: the Machine name
        @param unknownOnNotfound: if True, the method will return status UNKNOW for machines that doesn't exit. else, an excption will be thrown.  
        """
        vbox = getVirtualBox()
        try:
            machine = vbox.findMachine(name)
        except Exception, e:
            if unknownOnNotfound:
                return q.enumerators.VirtualboxMachineStatusType.UNKNOWN
            else:
                raise e
        
        return getMachineStateEnumFromStateNr(machine.state)
        
    def __waitTillMachineIsPoweredOff(self, machine, maxWaitTime):
        """
        This method waits on a VMachine trying to power it off.
        
        @param machine: IMachine instance.
        @param maxWaitTime: Maximum wait time in seconds. 
        """
        time.sleep(1)
        
        machineStopped = False
        machineState = self.__getMachineState(machine)
        starttime = time.time()
        
        # While we didn't pass the max. wait time and the machine is not powered off yet
        while time.time() - starttime < maxWaitTime and machineState != VirtualboxMachineStatusType.STOPPED:
            # Wait for little bit
            time.sleep(1)
            # Get recent machine state
            machineState = self.__getMachineState(machine)

        # If the VMachine is not powered off yet
        return False if machineState != VirtualboxMachineStatusType.STOPPED else True

    def __watchProgressStatus(self, progress, maxWaitTime):
        """
        This method watches the progress of executing long running commands against a VMachine.
        
        @param progress: IProgress instance which helps watching the progress of the running command process.
        @param maxWaitTime: Maximum wait time in milliseconds.  
        """
        progress.waitForCompletion(maxWaitTime)
        if progress.percent != self.___100_PERCENT_PROGRESS_ and progress.errorInfo is not None:
            raise RuntimeError(progress.errorInfo.text)

    def __getMachineState(self, machine):
        """
        A method to make a VMachine state number into and enumeration value.
        """
        return getMachineStateEnumFromStateNr(machine.state)

    def _getMachineObjByName(self, machineName):
        vbox = getVirtualBox()
        try:
            machine = vbox.findMachine(machineName)
            return machine
        except Exception, e:
            raise ValueError("Can't find machine %s" % machineName)

    def _openSession(self, machineName, requireLock=True):
        vbox = getVirtualBox()
        machine = self._getMachineObjByName(machineName)
        session = getSessionObject()
        if requireLock:
            machine.lockMachine(session, VirtualBoxLockType.SHARED.value)
        return session, machine
        
        q.eventhandler.raiseCriticalError('Could not get vbox session')

    def _closeSession(self, session):
        try:
            session.unlockMachine()
        except Exception, e:
            q.logger.log("Error while closing VBox session. Error: %s" %e, 3)