from pylabs import q
from pylabs.baseclasses.CommandWrapper import CommandWrapper
from VirtualboxEnums import VirtualboxMediumType
from VBoxTools import getVirtualBox, getVirtualBoxManager


class VirtualBoxTools(CommandWrapper):

    def createVmdk(self, devicePath, filePath):
        """
        Generates a vmdk file on filePath for the given device.

        @type devicePath: string
        @param devicePath: Path to the device to wrap in the VMDK

        @type filePath: string
        @param filePath: Path to the vmdk to generate
        """

        if not q.system.fs.exists(devicePath):
            raise ValueError('The specified device does not exist.')

        q.logger.log("Creating vmdk with path %s for device %s" % (filePath, devicePath), 3)
        command = "VBoxManage internalcommands createrawvmdk -filename %(filename)s -rawdisk %(diskname)s" % {'filename':filePath, 'diskname':devicePath}
        exitCode, output = q.system.process.execute(command, outputToStdout=True)
        if exitCode == 0:
            q.logger.log("VMDK %s created successfully" % filePath, 3)
        else:
            q.logger.log("Failed to create VMDK [%s]. Reason: %s"%(filePath, output), 3)
            raise RuntimeError("Failed to create VMDK [%s]"%filePath)


    def createVDI(self, filePath, vdiSize):
        """
        Generates a vdi file on filePath for the given device.

        @type devicePath: string
        @param devicePath: Path to the device to wrap in the VMDK

        @type filePath: string
        @param filePath: Path to the vdi to generate
        """
        q.logger.log("Creating VDI with path %s" % (filePath), 3)

        command = "VBoxManage createhd --filename %s -size %s -register" % (filePath, vdiSize)

        exitCode, output = q.system.process.execute(command, outputToStdout=True)
        if exitCode == 0:
            q.logger.log("VDI %s created successfully" % filePath, 3)
        else:
            q.logger.log("Failed to create VDI [%s]. Reason: %s"%(filePath, output), 3)
            raise RuntimeError("Failed to create VDI [%s]"%filePath)


    def listHdds(self):
        """
        Retrieve all available harddrives
        """
        q.logger.log("Retrieving all hdds", 3)
        
        command = "VBoxManage list hdds"
        
        exitCode, output = q.system.process.execute(command, outputToStdout=True)
        if exitCode == 0:
            q.logger.log("Retrieved all hdds successfully", 3)
            return output
        else:
            q.logger.log("Failed retrieving all hdds. Reason: %s"%(filePath, output), 3)
            raise RuntimeError("Failed to list hdds")


    def closeMedium(self, filePath, mediumType=VirtualboxMediumType.DISK):
        """
        Close the specified medium
        
        @param filePath: file name of the medium
        @type filePath: string
        
        @param mediumType: type of the medium (disk | dvd | floppy)
        @type mediumType: VirtualboxMediumType enum 
        """
        
        command = "VBoxManage closemedium %s %s" % (mediumType, filePath)
        
        exitCode, output = q.system.process.execute(command, outputToStdout=True)
        if exitCode == 0:
            q.logger.log("Successfully closed %s medium" % mediumType, 3)
        else:
            q.logger.log("Failed to close %s medium. Reason: %s"%(mediumType, output), 3)
            raise RuntimeError("Failed to close %s medium [%s]"%(mediumType, name))
        
    def listGuestProperties(self, machineName):
        """
        Return a list of the guest properties matching a set of patterns along
        with their values, time stamps and flags.
        
        @param machineName: machine name
        @type machineName: string
        
        @return: list of dicts [{'name':'', value:'', 'timestamp':'', 'flags':''}]
        """
        vbox = getVirtualBox()
        machine = vbox.findMachine(machineName)

        gp = machine.enumerateGuestProperties('')
        guestproperties = list()
        for x in xrange(len(gp[0])):
            guestproperties.append({'name':gp[0][x],
                                    'value':gp[1][x],
                                    'timestamp':gp[2][x],
                                    'flags':gp[3][x]})
        return guestproperties
    
    @staticmethod
    def listHostUsbDevices(USBDeviceState=None):
        """
        List of USB devices available on host
        If no USBDeviceState is specified, all usb devices will be returned
        
        @param USBDeviceState: USB device state
        @type USBDeviceState: USBDeviceState enum
        """
        vboxmg = getVirtualBoxManager()
        usbdevices = vboxmg.getArray(vboxmg.vbox.host, 'USBDevices')
        return [usbdev for usbdev in usbdevices if not USBDeviceState or usbdev.state == USBDeviceState.value]
    
    def addISCSIDisk(self,machineName, controllername, port, server, target):
        command = 'VBoxManage storageattach "%(machineName)s" --storagectl "%(controllername)s" --port %(port)s --device 0 --type hdd --medium iscsi --server %(server)s --target "%(target)s"'
        
        exitCode, output = q.system.process.execute(command % {'machineName'    : machineName,
                                                               'controllername' : controllername,
                                                               'port'           : port,
                                                               'server'         : server,
                                                               'target'         : target} , 
                                                    outputToStdout=True)
        if exitCode == 0:
            q.logger.log("Successfully added ISCSI disk %s to machine %s" % (target,machineName), 3)
        else:
            q.logger.log("Failed to add ISCSI disk %s to machine %s" % (target,machineName), 3)
            raise RuntimeError("Failed to add ISCSI disk %s to machine %s" % (target,machineName), 3)
    
    def createSerialPort(self, machineName, port, IObase, IRQ):
        """
        Create serial port on vmachine.
        
        @param machineName: machine name
        @type machineName: string
        
        @param port: serial port identifier
        @type port: int
        
        @param IObase:  I/O base address [COM1: I/O base 0x3F8, IRQ 4, COM2: I/O base 0x2F8, IRQ 3, COM3: I/O base 0x3E8, IRQ 4, COM4: I/O base 0x2E8, IRQ 3]
        @type IObase: string
        
        @param IRQ: interrupt
        @type IRQ: int
        """
        commands = ['VBoxManage modifyvm "%(machineName)s" --uart%(port)s %(iobase)s %(irq)s']
        commands.append('VBoxManage modifyvm "%(machineName)s" --uartmode%(port)s disconnected')
        
        
        for command in commands:
            exitCode, output = q.system.process.execute(command % {'machineName' : machineName,
                                                                   'port'        : port,
                                                                   'iobase'      : IObase,
                                                                   'irq'         : IRQ},
                                                        outputToStdout=True)
            if exitCode == 0:
                q.logger.log("Successfully added serial port %s to machine %s" % (port,machineName), 3)
            else:
                q.logger.log("Failed to add serial port %s to machine %s" % (port,machineName), 3)
                raise RuntimeError("Failed to add serial port %s to machine %s" % (port,machineName), 3)
    
    def removeSerialPort(self, machineName, port):
        """
        Create serial port on vmachine.
        
        @param machineName: machine name
        @type machineName: string
        
        @param port: serial port identifier
        @type port: int
        """
        command = 'VBoxManage modifyvm "%(machineName)s" --uart%(port)s off'
        
        exitCode, output = q.system.process.execute(command % {'machineName' : machineName,
                                                               'port'        : port},
                                                    outputToStdout=True)
        if exitCode == 0:
            q.logger.log("Successfully removed serial port %s from machine %s" % (port,machineName), 3)
        else:
            q.logger.log("Failed to removed serial port %s from machine %s" % (port,machineName), 3)
            raise RuntimeError("Failed to removed serial port %s from machine %s" % (port,machineName), 3)
        